//----------------------------------------------------------------------------
// 
//	file: spikelayer2.cc
// 
//	function: make spikelayer more robust to parameter change 
//	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
//	creation: Fri Nov 27 04:52:03 CST 1998
// 
//	$Revision: 1.4 $ $Date: 2001/11/30 20:24:15 $
//	$Log: spikelayer2.cc,v $
//	Revision 1.4  2001/11/30 20:24:15  yschoe
//	pglissom v1.0 release
//
//	Revision 1.3  2000/01/30 14:56:21  yschoe
//	*** empty log message ***
//
//	Revision 1.2  1999/12/22 05:29:48  yschoe
//	print_stat added
//
//	Revision 1.1  1999/02/09 07:42:25  yschoe
//	Initial revision
//
//
//----------------------------------------------------------------------------

#include "spikelayer2.h"

//----------------------------------------------------------------------------
//	spikelayer2() 
//	  - Constructor
//----------------------------------------------------------------------------
spikelayer2::spikelayer2(char *name, int cols, int rows, 
		 int exc_rad, int inh_rad,
                 int max_inp_layers, netlayer *feeding,
                 int inp_cols, int inp_rows, int  aff_rad,
                 REAL synapse_decay, REAL inhloop_decay, REAL theta, REAL tau,
		 int padded_flag)
	: spikelayer(name, cols, rows, exc_rad, inh_rad, max_inp_layers,
		 feeding, inp_cols, inp_rows, aff_rad, synapse_decay, 
		 inhloop_decay, theta, tau,padded_flag) {

  // use normalized LUTs (i.e., the leaky sums are divided by the number of
  //	bits)
  delete [] synapse_lut;
  delete [] inhloop_lut;

  synapse_decay_mult = exp(-synapse_decay);

  min_net_input = 0.0;
  max_net_input = 1.0;
  synapse_lut  = init_normalized_leaky_table(SPIKE_QLEN,synapse_decay);
  inhloop_lut  = init_normalized_leaky_table(SPIKE_QLEN,inhloop_decay);

  auto_theta_flag = 0;  // by default
  auto_theta_rate = 0.475;  // by default

}

//----------------------------------------------------------------------------
//	activate() 
//	  - activate spike layer
//----------------------------------------------------------------------------
void spikelayer2::activate(int calc_aff) {
  
  // 0. calc and push spikes
  push_spike();

  // 1. use netlayer::calc_affresp() to calculate net_input.
  //    todo: If we assume that input spikes reach a stable firing before
  //    feeding into the cortex, we can avoid calculating affresp 
  //    everytime during settling.
  if (calc_aff==0) {
  calc_affresp();       // net_aff_input <-- wgt * (feeding->output); 
  }

  // 2. use netlayer::calc_latresp() to calculate lateral contrib
  calc_latresp();       // net_exc/inh_input <- wgt*output (from previous step)
  calc_contrib_sum(0);  // net_input <-- net_aff_input + net_exc/inh_input

  // SPIKELAYER2 : pass net_input through sigmoid : output <- sigmoid(net_input)
  apply_actfunc();

  if (calc_aff==0 && auto_theta_flag == 1) {
    nonzero_minmax(output, &min_net_input, &max_net_input);
    // update_sig(min_net_input,max_net_input);
    //cout << "net_input_minmax " << min_net_input << " " << max_net_input << endl;
    theta = max_net_input * auto_theta_rate;
    cout << "new_theta = " << theta << endl;
    cerr << "new_theta = " << theta << endl;
  }

  // 3. calculate spike
  calc_spike();         // output <-- shifted, decayed sum of spikes

}

//----------------------------------------------------------------------------
//      calc_spike() 
//        - generate spike and push spike, put LUT value, etc.
//----------------------------------------------------------------------------
void spikelayer2::calc_spike() {

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

        // 0. just for displaying -- never used in calculation 
        lfp[i][j] = output[i][j];  // this was calculated in activate().

        // 1. generate spike
        // *. spikeact_q is identical to inhloop_q, so there's
        //    spikeact_q is used instead of inhloop_q when looking up the
        //    inhloop_lut.
        // spike[i][j] = ( output[i][j]
        spike[i][j] = ( output[i][j]
                > theta
                  +tau*inhloop_lut[(unsigned)spikeact_q[i][j] & SPIKE_QMASK]
		  +intrinsic_noise*drand48()
            ) ? 1 : 0;
        /* DEBUG
        if (i==2 && j==2) {
                disp_binary(spikeact_q[i][j] & SPIKE_HISTQMASK);
                cout << "  :  " 
                     << ((unsigned)spikeact_q[i][j] & SPIKE_HISTQMASK) ;
        }*/

     }
  }

}

//----------------------------------------------------------------------------
//      calc_spike_accumulate() 
//        - generate spike and push spike, put LUT value, etc.
//	  - accumulate past activity not leading to a spike
//----------------------------------------------------------------------------
void spikelayer2::calc_spike_accumulate() {

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

        // 0. just for displaying -- never used in calculation 
        lfp[i][j] = output[i][j];  // this was calculated in activate().

        // 1. generate spike
        // *. spikeact_q is identical to inhloop_q, so there's
        //    spikeact_q is used instead of inhloop_q when looking up the
        //    inhloop_lut.
        // spike[i][j] = ( output[i][j]
        spike[i][j] = ( output[i][j]
		  +prev_output[i][j]
                > theta
                  +tau*inhloop_lut[(unsigned)spikeact_q[i][j] & SPIKE_QMASK]
		  +intrinsic_noise*drand48()
            ) ? 1 : 0;
        /* DEBUG
        if (i==2 && j==2) {
                disp_binary(spikeact_q[i][j] & SPIKE_HISTQMASK);
                cout << "  :  " 
                     << ((unsigned)spikeact_q[i][j] & SPIKE_HISTQMASK) ;
        }*/
	if (spike[i][j]) {
		prev_output[i][j] *= synapse_decay_mult;
	} else {
		prev_output[i][j] = synapse_decay_mult 
		/*
				* (prev_output[i][j] + net_aff_input[i][j]);
		*/
			*( prev_output[i][j]
			+ net_aff_input[i][j] 
			+ lat_exc_contri*net_exc_input[i][j]
			- lat_inh_contri*net_inh_input[i][j]);

		// + output[i][j];
	}

     }
  }

}

//----------------------------------------------------------------------------
//      print_stat() 
//        - print status of (col,row) neuron.
//----------------------------------------------------------------------------
void spikelayer2::print_stat(int col, int row) {

  cout  << net_aff_input[col][row] << " "
        << (lat_exc_contri*net_exc_input[col][row]) << " "
        << (lat_inh_contri*net_inh_input[col][row]) << " "
        << act_func(net_input[col][row]) << " "
        << (synapse_lut[(unsigned)spikeact_q[col][row]&SPIKE_QMASK ]) << " "
        << theta << " "
        << (output[col][row]-tau*inhloop_lut[(unsigned)spikeact_q[col][row]&SPIKE_QMASK])<< " " << col << " " << row << endl;

}
//----------------------------------------------------------------------------
void spikelayer2::print_stat(int cur_inp, int inp_col, int inp_row) {

  int col = (int)(inp_col * cols / (REAL)feedings[cur_inp]->cols);
  int row = (int)(inp_row * rows / (REAL)feedings[cur_inp]->rows);

  cout  << net_aff_input[col][row] << " "
        << (lat_exc_contri*net_exc_input[col][row]) << " "
        << (lat_inh_contri*net_inh_input[col][row]) << " "
        << act_func(net_input[col][row]) << " "
        << (synapse_lut[(unsigned)spikeact_q[col][row]&SPIKE_QMASK ]) << " "
        << theta << " "
        << (output[col][row]-tau*inhloop_lut[(unsigned)spikeact_q[col][row]&SPIKE_QMASK])<< " " << col << " " << row << endl;

}


//----------------------------------------------------------------------------
//	kill_recursive_lat() 
//	  - kill recursive lateral connections
//----------------------------------------------------------------------------
void spikelayer2::kill_recursive_lat(){

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

        // 1.1 excitatory
        int lowx = lat_exc_range[i][j].lowx;
        int lowy = lat_exc_range[i][j].lowy;
	wgt[i][j].lat_exc_w[i-lowx][j-lowy] = -DEATH_FLAG;

        // 1.2 inhibitory
        lowx = lat_inh_range[i][j].lowx;
        lowy = lat_inh_range[i][j].lowy;
        wgt[i][j].lat_inh_w[i-lowx][j-lowy] = -DEATH_FLAG;

     }
  }

}

